# (C) 2001-2024 Intel Corporation. All rights reserved.
# Your use of Intel Corporation's design tools, logic functions and other 
# software and tools, and its AMPP partner logic functions, and any output 
# files from any of the foregoing (including device programming or simulation 
# files), and any associated documentation or information are expressly subject 
# to the terms and conditions of the Intel Program License Subscription 
# Agreement, Intel FPGA IP License Agreement, or other applicable 
# license agreement, including, without limitation, that your use is for the 
# sole purpose of programming logic devices manufactured by Intel and sold by 
# Intel or its authorized distributors.  Please refer to the applicable 
# agreement for further details.

# #####################################################################
# Swizzling Tool - Software Entry Point
# #####################################################################
# VERSION
# 1.0.0
#
# FILE DESCRIPTION
# ----------------
# This script processes `.fir.rpt` and `.ip` files from a compiled Quartus Prime Pro project
# that includes an EMIF IP. It generates a CSV file listing the EMIF signal names and their
# locations. Users can then specify their desired pin swizzling directly within this CSV file.
# Subsequently, the tool is capable of producing the appropriate string to be inserted into
# the Advanced Parameters tab of the EMIF IP Parameter Editor.
#
# DISCLAIMERS
# ----------------
# The Swizzling Tool is currently in beta. Its features and software interfaces may undergo
# changes in future updates.
#
# USAGE
# ----------------
# To utilize the tool, place the Swizzling Tool directory at the root of your project directory.
#
# Creating the Pin Swizzle CSV File:
# Execute the following command to generate the `pin_swizzle.csv` file:
#
# quartus_py swizzling_tool/main.py --ip-file <path to your emif .ip file> --fitter-report <path to your fit.rpt> --interface-pattern <EMIF signal naming pattern>
#
# Example:
# quartus_py swizzling_tool/main.py --ip-file ./ip/ed_synth_emif_ph2_0.ip --fitter-report ./ed_synth.fit.rpt --interface-pattern emif_ph2_0
#
# Generating the Swizzling String:
# To create the swizzling string after defining your pin swizzle in the CSV file, use the following command:
#
# quartus_py swizzling_tool/main.py --pin-swizzle
#
# Please ensure to complete the command with the necessary arguments as per your project's requirements.

from collections import OrderedDict
import argparse
import json
import csv
import sys
from parse_emif_ip import tester_function, get_basic_ip_info

parser =  argparse.ArgumentParser(description='Parse fitter report and pin file to get pins of External Memory Interfaces.')
parser.add_argument('--test', help='Test help')
parser.add_argument('--interface-pattern', help='Interface pattern to search for')
parser.add_argument('--fitter-report', help='Fitter report path')
parser.add_argument('--ip-file', help="EMIF IP file path")
parser.add_argument('--pin-swizzle', help='Perform swizzle ', action='store_true')

args =  parser.parse_args()
bank_list = []
working_csv_file_path = './swizzling_tool/pin_swizzle.csv'
json_scheme = './swizzling_tool/schemes.json'

def save_parameters_from_ip_file(emif_ip_file_path):
    info_dict = get_basic_ip_info(emif_ip_file_path)
    lines_2_write = [f"{line}={info_dict[line]}\n" for line in info_dict]
    with open('./swizzling_tool/.ip_parameters','w') as ip_parameters_file:
        ip_parameters_file.writelines(lines_2_write)

def parse_parameters_from_ip_file():
    read_parameters_list = []
    read_parameters_dict = {}
    with open('./swizzling_tool/.ip_parameters','r') as ip_parameters_file:
        read_parameters_list=ip_parameters_file.readlines()
    for line in read_parameters_list:
        line = line.strip(' \n').split('=')
        read_parameters_dict[line[0]] = line[1]
    return read_parameters_dict

def select_scheme(ip_parameters,bank_list,bank_dict,ac_pin,index_bl_addressed):
    mem_technology=ip_parameters['MEM_TECHNOLOGY'].split("_")[-1]
    mem_format = ip_parameters['MEM_FORMAT']
    ac_lanes = ip_parameters['USER_MIN_NUM_AC_LANES']
    ecc_mode = ip_parameters['CTRL_ECC_MODE']
    total_dq=int(ip_parameters['MEM_DEVICE_DQ_WIDTH'])
    
    if mem_technology == "DDR4":
        if mem_format == "MEM_FORMAT_DISCRETE" and total_dq >= 16 and total_dq <= 40:
            
            if total_dq == 16:
                if ac_lanes == "3":
                    return f"{mem_technology}_3ac_x{total_dq}"
                elif ac_lanes == "4":
                    return f"{mem_technology}_4ac_x{total_dq}"
            elif total_dq == 24 and ecc_mode == "CTRL_ECC_MODE_ENABLED":
                if ac_lanes == "3":
                    return f"{mem_technology}_3ac_x16_ecc"
                elif ac_lanes == "4":
                    return f"{mem_technology}_4ac_x16_ecc"
            elif total_dq == 32:
                if ac_lanes == "3":
                    return f"{mem_technology}_3ac_x{total_dq}"
                elif ac_lanes == "4":
                    return f"{mem_technology}_4ac_x{total_dq}"
            elif total_dq == 40 and ecc_mode == "CTRL_ECC_MODE_SIDE_BAND" :
                return f"{mem_technology}_x32_ecc"
            
            else:
                print("No supported scheme for discrete device")
                sys.exit()

        elif total_dq ==64 or total_dq == 72:
            ac_pin_index = int(ac_pin['index_within_bytelane'])
            ac_position=""
            if ac_pin_index>=0 and ac_pin_index <= 47:
                ac_position="bot"
            elif ac_pin_index>=48 and ac_pin_index <= 95:
                ac_position="top"
            
            for index_w_bl_sec_ctrll in index_bl_addressed[bank_list[0]]:
                if int(index_w_bl_sec_ctrll) >=0 and int(index_w_bl_sec_ctrll) <= 47:
                    sec_dq_pos="bot"
                elif int(index_w_bl_sec_ctrll) >=48 and int(index_w_bl_sec_ctrll) <= 95:
                    sec_dq_pos="top"
                break
            return f"{mem_technology}_x{total_dq}_acpri{ac_position}_dqsec{sec_dq_pos}"
        else :
            print("Memory scheme not suported")
            sys.exit()

    elif mem_technology == "DDR5":
        if mem_format == "MEM_FORMAT_DISCRETE" and total_dq >= 16 and total_dq <= 40:
            ac_pin_index = int(ac_pin['index_within_bytelane'])
            ac_position=""
            if ac_pin_index>=0 and ac_pin_index <= 47:
                ac_position="bot"
                if total_dq == 16:
                    return f"{mem_technology}_ac{ac_position}_x{total_dq}"
            elif ac_pin_index>=48 and ac_pin_index <= 95:
                ac_position="top"
                if total_dq == 16:
                    return f"{mem_technology}_ac{ac_position}_x{total_dq}"
            else :
                print("Memory scheme not suported")
                sys.exit()
            


    else:
        print("Memory scheme not suported")
        print("The EMIF memory scheme you are using will be supported in an upcoming version, please contact your sales representative to get more information")
        sys.exit()

def parse_fir_rpt (initial_pin_dict,interface_pattern):
    #print(initial_pin_dict)
    in_range = False

    highest_index = 0
    current_bank = ""
    bank_list = []
    with open(fit_report_path,"r") as fitt_file:
        end_pattern_counter = 0
        for line in fitt_file:
            if line.startswith("; Detailed I/O Block Info") and in_range==False:
                in_range=True
                
            # Pattern "+----------" is used to delimit tables in the fitter report
            # after seeing a line starting with "; Detailed I/O Block Info" it will appear 
            # 3 times, the last marks the end of the table.
            if in_range and line.startswith("+----------") :
                end_pattern_counter += 1
                if end_pattern_counter == 3:
                    break

            if in_range:
                pin_info = [word.strip(' \n') for word in line.split(";")]
                if len(pin_info)>=2:
                    current_bank = pin_info [1] if pin_info[1] != "--" else current_bank
                if len(pin_info)>=5:
                    
                    if pin_info[4].startswith(interface_pattern):
                        pin = pin_info[3]
                        io96_index_within_bank = pin_info[2]
                        signal_name = pin_info[4]
                        
                        initial_pin_dict[signal_name] = {'pin' : pin,'bank':current_bank, 'bank_index':io96_index_within_bank} 
                        bank = current_bank[0:2]
                        if bank not in bank_list:
                            bank_list.append(bank)

        
        json_version = json.dumps(initial_pin_dict)
        
        return initial_pin_dict, bank_list
    ###
def populate_bl_dict(bl_dict,initial_pin_dict):
    for element in initial_pin_dict:
        bank =  initial_pin_dict[element]['bank'][0:2]
        index = int(initial_pin_dict[element]['bank_index'])
        
        if index >= 0 and index <= 11 :
            bl_dict[bank]['BL0'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 12 and index <= 23 :
            bl_dict[bank]['BL1'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 24 and index <= 35 :
            bl_dict[bank]['BL2'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 36 and index <= 47 :  
            bl_dict[bank]['BL3'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 48 and index <= 59 :
            bl_dict[bank]['BL4'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 60 and index <= 71 :
            bl_dict[bank]['BL5'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 72 and index <= 83 :
            bl_dict[bank]['BL6'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}
        elif index >= 84 and index <= 95 :
            bl_dict[bank]['BL7'][index] = {'name': element, 'pin': initial_pin_dict[element]['pin']}     
    return bl_dict

def write_emif_signals_to_csv(bl_dict):

    with open(working_csv_file_path, mode='w') as csv_file:
        fieldnames = ['bank','Byte Lane','Index within BL', 'name','pin','pin swizzle']
        csv_writer = csv.DictWriter(csv_file,fieldnames=fieldnames)
        csv_writer.writeheader()

        for bank in bl_dict:
            for bl_index in range(0,8):
                for index in range(12*bl_index,12*bl_index+12) :
                    try:
                        csv_writer.writerow({'bank':bank,'Byte Lane':f'BL{bl_index}','Index within BL':index, 'name':bl_dict[bank][f'BL{bl_index}'][index]['name'],'pin': bl_dict[bank][f'BL{bl_index}'][index]['pin'], 'pin swizzle':''})
                    except KeyError:
                        pass
                        

if __name__ == "__main__":
    #Select program flow based on args and flags
    fit_report_path = args.fitter_report
    interface_pattern = args.interface_pattern
    ip_file_path = args.ip_file 
    pin_swizzle = args.pin_swizzle

    procedure = ""

    if fit_report_path and interface_pattern and ip_file_path and (not pin_swizzle):

        procedure = "create_csv"

        #Create pin dictionary, addressed by the name of signals
        initial_pin_dict = {}
        initial_pin_dict, bank_list = parse_fir_rpt(initial_pin_dict,interface_pattern)

        #Initialize byte_lane dictionary
        bl_dict = {}
        for bank in bank_list:
            bl_dict[bank] = {}
            for bl_index in range(0,9):
                byte_lane =  f'BL{bl_index}'
                bl_dict[bank][byte_lane] = {}

        #Populate byte_lane dict
        bl_dict = populate_bl_dict(bl_dict,initial_pin_dict)

        write_emif_signals_to_csv(bl_dict)

        save_parameters_from_ip_file(ip_file_path)
        print("Signal names, bank, and pin locations of your EMIF IP have been saved to the pin_swizzle.csv file. Please open it and perform the desired pin swizzle.")

    elif not fit_report_path and not interface_pattern and not ip_file_path and (pin_swizzle):
        procedure = "perform_swizzle"
        schemes_dict = {}
        index_bl_addressed = {}
        assigned_pin_addressed = {}
        swizzle_string = []
        ac_pin = {}
        bank_list = []
        bank_dict = {}
        inv_bank_dict = {}

        with open(json_scheme, mode='r') as json_schemes_file :
            schemes_dict = json.load(json_schemes_file)
            
        with open(working_csv_file_path, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file)
            for row in csv_reader:
                if row['bank'] in index_bl_addressed:
                    index_bl_addressed[row['bank']].update({row['Index within BL']:{'byte_lane':row['Byte Lane'],'name':row['name'],'pin':row['pin'],'pin_swizzle':row['pin swizzle']}})
                    assigned_pin_addressed[row['bank']].update({row['pin']:{'byte_lane':row['Byte Lane'],'name':row['name'],'index_within_bytelane':row['Index within BL'],'pin_swizzle':row['pin swizzle']}})
                    if row['pin swizzle'] == "*" :
                        ac_pin = {'bank': row['bank'], 'byte_lane':row['Byte Lane'],'pin':row['pin'],'index_within_bytelane':row['Index within BL']}

                else:
                    bank_list.append(row['bank'])
                    bank_dict[row['bank']]=""
                    index_bl_addressed[row['bank']] = {}
                    assigned_pin_addressed[row['bank']] = {}
                    index_bl_addressed[row['bank']].update({row['Index within BL']:{'byte_lane':row['Byte Lane'],'name':row['name'],'pin':row['pin'],'pin_swizzle':row['pin swizzle']}})
                    assigned_pin_addressed[row['bank']].update({row['pin']:{'byte_lane':row['Byte Lane'],'name':row['name'],'index_within_bytelane':row['Index within BL'],'pin_swizzle':row['pin swizzle']}})
        
        #inv_bank_dict "pri" key stores the bank that cointains AC lanes
        if 'bank' in ac_pin:
            bank_dict[ac_pin['bank']] = "primary"
            inv_bank_dict['pri'] = ac_pin['bank']
            if len(bank_list) > 1:
                bank_list.remove(ac_pin['bank'])
                bank_dict[bank_list[0]] = "secondary"
                inv_bank_dict['sec'] = bank_list[0]
        else:
            print("No address or command pin was specified in the csv file, please indicate any AC pin using a \"*\" in the pin swizzle column next to the selected AC pin")
            sys.exit()

        ip_parameters = parse_parameters_from_ip_file()
        scheme = select_scheme(ip_parameters,bank_list,bank_dict,ac_pin,index_bl_addressed).lower()

        
        total_dq=int(ip_parameters['MEM_DEVICE_DQ_WIDTH'])
        mem_technology=ip_parameters['MEM_TECHNOLOGY'].split("_")[-1]

        print(scheme)
        if mem_technology == "DDR4" or mem_technology == "DDR5":
            if len(schemes_dict[scheme+'_dq'])==1:
                lockstep_flag = False
            elif len(schemes_dict[scheme+'_dq'])==2:
                lockstep_flag = True 

            global_swizzling_dict = {}
            for controller in schemes_dict[scheme+'_dq']:
                
                for dq_index in schemes_dict[scheme+'_dq'][controller]:
                    
                    dq_pin_to_swizzle = index_bl_addressed[inv_bank_dict[controller[0:3]]][ schemes_dict[scheme+'_dq'][controller][dq_index] ]['pin_swizzle']
                    if dq_pin_to_swizzle != "" :
                        global_swizzling_dict[dq_index]=[schemes_dict[scheme+'_index_within_bl'][controller][assigned_pin_addressed[inv_bank_dict[controller[0:3]]][dq_pin_to_swizzle]['index_within_bytelane']], controller[0:3]]
                    else:
                        global_swizzling_dict[dq_index]=[dq_index,controller[0:3]]
                        
            final_string = ""  
            for dqs_index in range(0,int(total_dq/8)):
                if lockstep_flag == False:
                    aux_swizzle_string = f"PIN_SWIZZLE_CH0_DQS{dqs_index}="
                elif lockstep_flag == True:
                    current_controller = global_swizzling_dict[str(dqs_index*8)][1]
                    aux_swizzle_string = f"PIN_SWIZZLE_{current_controller.upper()}_DQS{dqs_index}="
                changes_in_dqs_group = 0 
                
                for dq_index in range(8*dqs_index,8*dqs_index+8):#This line to change due to x4 and x16 devices
                    if str(dq_index) != str(global_swizzling_dict[str(dq_index)][0]):
                        changes_in_dqs_group = changes_in_dqs_group + 1
                    if dq_index<8*dqs_index+7:
                        aux_swizzle_string = aux_swizzle_string+f"{global_swizzling_dict[str(dq_index)][0]},"
                    else:
                        aux_swizzle_string = aux_swizzle_string+f"{global_swizzling_dict[str(dq_index)][0]}"
                if changes_in_dqs_group>0:
                    final_string = final_string+aux_swizzle_string+';'
            print(final_string)
    else :
        print("Can't parse arguments, --pin-swizzle must be used alone")
        sys.exit()
